// ======================================================================
//
// gl_wu2.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "first.h"
#include "gl_wu2.h"

#include "DirectDraw.h"
#include "gl_init.h"
#include "Geometry.h"
#include "Material.h"
#include "MaterialData_wu2.h"
#include "Palette.h"
#include "PaletteData_wu2.h"
#include "PixelIndex.h"

#include "clut.h"
#include "raster.h"
#include "sbuf.h"

#include <assert.h>
#include <stdio.h>


#pragma optimize("", off)


// ======================================================================

#define TRACE(a)          NOP // DPRINT_E_TRACE(a)
#define DPRINT_F(a)       NOP // DPRINT_E_JG(a)
#define RECORD(a)         NOP // DPRINT_LOG(a)

// ======================================================================

const int RASTER_VERTEX_CNT = 16;

// ======================================================================

// -q- i would have liked to have put these in the class header, but
// doing it here means that I don't have to include david's header files
// in gl_wu2.h.  Oh well, at least I made them static.

static clut_miser::color_table          colorTable;
static clut_miser::pixel_blend_table_8  blendTable;
static clut_miser::output_settings      settings;
static clut_miser::generator           *clut_generator;

MaterialData_wu2       *GL_wu2::wirefill;
const MaterialData_wu2 *GL_wu2::material;
bool                    GL_wu2::in3dRender;
float                   GL_wu2::vx0;
float                   GL_wu2::vy0;
float                   GL_wu2::vx1;
float                   GL_wu2::vy1;

// ======================================================================

inline void set_display_16(const clut_miser::pixel_format_16 & desc )
{
	RECORD(("raster::set_display_16(...);\n"));
	raster::set_display_16(desc);
}

// ----------------------------------------------------------------------

inline void set_edge_clipping(int do_clipping)
{
	RECORD(("raster::set_edge_clipping(%d);\n", do_clipping));
	raster::set_edge_clipping(do_clipping);
}

// ----------------------------------------------------------------------

inline void set_transparency_index(int index)
{
	RECORD(("raster::set_transparency_index(%d);\n", index));
	raster::set_transparency_index(index);
}

// ----------------------------------------------------------------------

inline void set_surface(int x_res, int y_res, void * buffer, u32 buf_pitch)
{
	RECORD(("raster::set_surface(%d, %d, ((void *)%p), %d);\n", x_res, y_res, buffer, buf_pitch));
	raster::set_surface(x_res, y_res, buffer, buf_pitch);
}

// ----------------------------------------------------------------------

inline void sweep_and_draw(void)
{
	RECORD(("raster::sweep_and_draw();\n"));
	raster::sweep_and_draw();
}

// ----------------------------------------------------------------------

inline void draw_transparent_polys(void)
{
	RECORD(("raster::draw_transparent_polys();\n"));
	raster::draw_transparent_polys();
}

// ----------------------------------------------------------------------

inline void submit_edge(float x0, float y0, float x1, float y1, raster::poly *p0, raster::poly *p1)
{
	RECORD(("raster::submit_edge(%9.5f, %9.5f, %9.5f, %9.5f, %p, %p);\n", x0, y0, x1, y1, p0, p1));
	raster::submit_edge(x0, y0, x1, y1, p0, p1);
}

// ----------------------------------------------------------------------

inline void dumpVertex(const char *start, const raster::vert_3 &v)
{
	UNREF(start);
	UNREF(v);
	RECORD(("  // %s c[0]=%9.5f  c[0]=%9.5f  z_inv=%9.5f\n", start, v.c[0], v.c[1], v.z_inv));
}

// ----------------------------------------------------------------------

inline void dumpVertex(const char *start, const raster::vert_3s &v)
{
	UNREF(start);
	UNREF(v);
	RECORD(("  // %s c[0]=%9.5f  c[0]=%9.5f  z_inv=%9.5f  shade=%9.5f\n", start, v.c[0], v.c[1], v.z_inv, v.shade));
}

// ----------------------------------------------------------------------

inline raster::poly *submit_als(const raster::vert_3s &v0, const raster::vert_3s &v1, const raster::vert_3s &v2, const raster::solid_surface_desc *sd)
{
	raster::poly *p;
	RECORD(("raster::submit_als();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	p = raster::submit_als(v0, v1, v2, sd);
	RECORD(("  // result=%p\n", p));
	return p;
}

// ----------------------------------------------------------------------

inline raster::poly *submit_pls(const raster::vert_3s &v0, const raster::vert_3s &v1, const raster::vert_3s &v2, const raster::solid_surface_desc *sd)
{
	raster::poly *p;
	RECORD(("raster::submit_pls();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	p = raster::submit_pls(v0, v1, v2, sd);
	RECORD(("  // result=%p\n", p));
	return p;
}

// ----------------------------------------------------------------------

inline raster::poly *submit_afs(const raster::vert_3 &v0, const raster::vert_3 &v1, const raster::vert_3 &v2, float shade, const raster::solid_surface_desc *sd)
{
	raster::poly *p;
	RECORD(("raster::submit_afs();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	RECORD(("  // shade=%9.5f\n", shade));
	p = raster::submit_afs(v0, v1, v2, shade, sd);
	RECORD(("  // result=%p\n", p));
	return p;
}

// ----------------------------------------------------------------------

inline raster::poly *submit_aft( const raster::vert_3 &v0, const raster::vert_3 &v1, const raster::vert_3 &v2, float shade, const raster::texture *tex, const float uv_0[2], const float uv_1[2], const float uv_2[2])
{
	raster::poly *p;
	RECORD(("raster::submit_aft();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	RECORD(("  // shade=%9.5f\n", shade));
	p = raster::submit_aft(v0, v1, v2, shade, tex, uv_0, uv_1, uv_2);
	RECORD(("  // result=%p\n", p));
	return p;
}

// ----------------------------------------------------------------------

inline raster::poly *submit_pft( const raster::vert_3 &v0, const raster::vert_3 &v1, const raster::vert_3 &v2, float shade, const raster::texture *tex, const float uv_0[2], const float uv_1[2], const float uv_2[2])
{
	raster::poly *p;
	RECORD(("raster::submit_pft();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	RECORD(("  // shade=%9.5f\n", shade));
	p = raster::submit_pft(v0, v1, v2, shade, tex, uv_0, uv_1, uv_2);
	RECORD(("  // result=%p\n", p));
	return p;
}

// ----------------------------------------------------------------------

inline raster::poly *submit_alt(const raster::vert_3s &v0, const raster::vert_3s &v1, const raster::vert_3s &v2, const raster::texture *tex, const float uv_0[2], const float uv_1[2], const float uv_2[2])
{
	raster::poly *p;
	RECORD(("raster::submit_alt();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	p = raster::submit_alt(v0, v1, v2, tex, uv_0, uv_1, uv_2);
	RECORD(("  // result=%p\n", p));
	return p;
}

// ----------------------------------------------------------------------

inline raster::poly *submit_plt(const raster::vert_3s &v0, const raster::vert_3s &v1, const raster::vert_3s &v2, const raster::texture *tex, const float uv_0[2], const float uv_1[2], const float uv_2[2])
{
	raster::poly *p;
	RECORD(("raster::submit_plt();\n", p));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	p = raster::submit_plt(v0, v1, v2, tex, uv_0, uv_1, uv_2);
	RECORD(("  // result=%p\n", p));
	return p;
}

// ----------------------------------------------------------------------

inline void submit_aft_tp( const raster::vert_3 &v0, const raster::vert_3 &v1, const raster::vert_3 &v2, const float shade, const raster::texture *tex, const float uv_0[2], const float uv_1[2], const float uv_2[2], const raster::vert_3 **all_verts, int num_verts)
{
	RECORD(("raster::submit_aft_tp();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	RECORD(("  // shade=%9.5f\n", shade));
	raster::submit_aft_tp(v0, v1, v2, shade, tex, uv_0, uv_1, uv_2, all_verts, num_verts);
}

// ----------------------------------------------------------------------

inline void submit_pft_tp( const raster::vert_3 &v0, const raster::vert_3 &v1, const raster::vert_3 &v2, const float shade, const raster::texture *tex, const float uv_0[2], const float uv_1[2], const float uv_2[2], const raster::vert_3 **all_verts, int num_verts)
{
	RECORD(("raster::submit_pft_tp();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	RECORD(("  // shade=%9.5f\n", shade));
	raster::submit_pft_tp(v0, v1, v2, shade, tex, uv_0, uv_1, uv_2, all_verts, num_verts);
}

// ----------------------------------------------------------------------

inline void submit_alt_tp(const raster::vert_3s &v0, const raster::vert_3s &v1, const raster::vert_3s &v2, const raster::texture *tex, const float uv_0[2], const float uv_1[2], const float uv_2[2], const raster::vert_3 **all_verts, int num_verts)
{
	RECORD(("raster::submit_alt_tp();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	raster::submit_alt_tp(v0, v1, v2, tex, uv_0, uv_1, uv_2, all_verts, num_verts);
}

// ----------------------------------------------------------------------

inline void submit_plt_tp(const raster::vert_3s &v0, const raster::vert_3s &v1, const raster::vert_3s &v2, const raster::texture *tex, const float uv_0[2], const float uv_1[2], const float uv_2[2], const raster::vert_3 **all_verts, int num_verts)
{
	RECORD(("raster::submit_plt_tp();\n"));
	dumpVertex("v0", v0);
	dumpVertex("v1", v1);
	dumpVertex("v2", v2);
	raster::submit_plt_tp(v0, v1, v2, tex, uv_0, uv_1, uv_2, all_verts, num_verts);
}

// ----------------------------------------------------------------------

#if 0
static void dump(const void *data, int length)
{
	int     i;
	const u_int8 *d8;

	d8 = reinterpret_cast<const u_int8 *>(data);
	for (i = 0; i < length; ++i)
	{
		if ((i & 15) != 0)
			DPRINT_LOG((", "));
		else
			DPRINT_LOG(("  "));

		DPRINT_LOG(("0x%02x", d8[i]));

		if ((i & 15) == 15)
			DPRINT_LOG(("\n"));
	}
}
#endif

// ======================================================================

GL_wu2::GL_wu2(void)
{
	Geometry::install(this);

	submitPointListFunction     = passPointList;
	submitLineListFunction      = passLineList;
	submitLineStripFunction     = passLineStrip;
	submitTriangleListFunction  = passTriangleList;
	submitTriangleStripFunction = passTriangleStrip;
	submitTriangleFanFunction   = passTriangleFan;

	wirefill = NULL;
}

// ----------------------------------------------------------------------

GL_wu2::~GL_wu2(void)
{
	if (wirefill)
	{
		materialDestroyData(wirefill);
		wirefill = NULL;
	}

	delete clut_generator;
	clut_generator = NULL;

	Geometry::remove();
}

// ----------------------------------------------------------------------

void GL_wu2::getInformation(Information *information)
{
	GL_abstract::getInformation(information);

	information->name         = "Software";
	information->versionMinor = 0;
	information->ranking      = 0;
}

// ----------------------------------------------------------------------

bool GL_wu2::query(int)
{
	return true;
}

// ----------------------------------------------------------------------

bool GL_wu2::init(const GL_input *input, GL_output *output)
{
	if (!GL_dd::init(input, output))
		return false;

	// create the direct draw surfaces
	DirectDraw::tripleBuffer();
	GL_dd::createSurfaces(input, output);

	// setup the clut generator
	settings.default_shade_bits = getShadeBits();
	settings.diffuse_scale      = 1.0f;
	settings.specular_start     = 1.0f;
	settings.specular_scale     = 1.0f;

	// setup the rasterizer color information
	switch (input->depth)
	{
		case 2:
		{
			clut_miser::pixel_format_16 pixelFormat16(DirectDraw::getRBitCount(), DirectDraw::getGBitCount(), DirectDraw::getBBitCount(), DirectDraw::getRPlaceShift(), DirectDraw::getGPlaceShift(), DirectDraw::getBPlaceShift());
			clut_generator = NEW clut_miser::generator_16(pixelFormat16);
			clut_generator->settings = settings;
			set_display_16(pixelFormat16);
		}
		break;

		default:
			DEBUG_FATAL(1, ("GL_wu2::init bad depth %d", input->depth));
	}

	// setup the rasterizer
	set_transparency_index(0);
	set_edge_clipping(1);

	// create a material to wireframe
	wirefill = NEW MaterialData_wu2(255, 255, 255);

	return true;
}

// ----------------------------------------------------------------------

void GL_wu2::setActive(bool)
{
}

// ----------------------------------------------------------------------

void GL_wu2::debugDump(void)
{
}

// ----------------------------------------------------------------------

bool GL_wu2::lock(void)
{
	static int frame;

	++frame;
	RECORD(("// frame %d\n", frame));

	// check if there is a background color to clear to
	if (backgroundMaterial)
	{
		// get the constant RGB colors
		int r = backgroundMaterial->getR();
		int g = backgroundMaterial->getG();
		int b = backgroundMaterial->getB();

		// clear the viewport using the constant RGB colors
		DirectDraw::clear(DirectDraw::makeLFB(r, g, b));
	}

	if (!lockDD())
		return false;

	set_surface(width, height, reinterpret_cast<Pixel16 *>(DirectDraw::screenPtr()), DirectDraw::screenPitch());

	return true;
}

// ----------------------------------------------------------------------

void GL_wu2::unlock(bool display)
{
	GL_dd::unlock(display);
}

// ----------------------------------------------------------------------

void GL_wu2::makeClut(const RGB *rgb, int numColors, byte **mem, void **clut)
{
	int                      i, size, alignMask;
	clut_miser::color_table  colorTable;

	// build color lookup table
	colorTable.num_colors = numColors;
	for (i = 0; i < numColors; ++i, ++rgb)
	{
		colorTable.entries[i][3] = 0;
		colorTable.entries[i][2] = rgb->r;
		colorTable.entries[i][1] = rgb->g;
		colorTable.entries[i][0] = rgb->b;
	}

	// get the size the clut will need to be
	size = clut_generator->calculate_clut_size(colorTable);

	// allocate enough memory to align the clut
	alignMask = clut_generator->suggest_clut_alignment() - 1;
	*mem = NEW byte[size + alignMask];

	// align the clut
	*clut = reinterpret_cast<void *>((reinterpret_cast<int>(*mem + alignMask)) & (~alignMask));

	// finally, make the clut
	clut_generator->make_clut(*clut, colorTable);
}

// ----------------------------------------------------------------------

PaletteData *GL_wu2::paletteCreateData(const Palette *pal)
{
	return NEW PaletteData_wu2(pal);
}

// ----------------------------------------------------------------------

void GL_wu2::paletteDestroyData(PaletteData *pal)
{
	delete pal;
}

// ----------------------------------------------------------------------

MaterialData *GL_wu2::materialCreateData(const StaticMaterial *mat)
{
	return NEW MaterialData_wu2(mat);
}

// ----------------------------------------------------------------------

void GL_wu2::materialDestroyData(MaterialData *mat)
{
	delete mat;
}

// ----------------------------------------------------------------------

void GL_wu2::beginRender(const Camera &camera)
{
	int x0, y0, x1, y1;

	// let the geometry subsystem know about the new camera
	Geometry::setCamera(camera);

	// get the camera viewport
	camera.getViewport(&x0, &y0, &x1, &y1);

	// setup the viewport as necessary for david's rasterizer
	vx0 = real(x0) + CONST_REAL(0.5);
	vy0 = real(y0) + CONST_REAL(0.5);
	vx1 = real(x1) + CONST_REAL(0.5);
	vy1 = real(y1) + CONST_REAL(0.5);

	// setup the surface to clip to the viewport
	set_surface(x1-x0+1, y1-y0+1, pixelIndex(reinterpret_cast<Pixel16 *>(DirectDraw::screenPtr()), x0, y0, DirectDraw::screenPitch()), DirectDraw::screenPitch());

	in3dRender = true;
}

// ----------------------------------------------------------------------

void GL_wu2::endRender(void)
{
	in3dRender = false;
	sweep_and_draw();
	draw_transparent_polys();

	vx0 = CONST_REAL(0.5);
	vy0 = CONST_REAL(0.5);
	vx1 = real(width) - CONST_REAL(0.5);
	vy1 = real(height) - CONST_REAL(0.5);

	set_surface(width, height, reinterpret_cast<Pixel16 *>(DirectDraw::screenPtr()), DirectDraw::screenPitch());
}

// ----------------------------------------------------------------------

void GL_wu2::setPrimitiveProcessing(PrimitiveProcessing)
{
}

// ----------------------------------------------------------------------

void GL_wu2::setObjectToWorld(const Xform &objectToWorld)
{
	Geometry::setObjectToWorld(objectToWorld);
}

// ----------------------------------------------------------------------

void GL_wu2::setLightingEnabled(bool lightingEnabled)
{
	Geometry::setLightingEnabled(lightingEnabled);
}

// ----------------------------------------------------------------------

void GL_wu2::setLightList(Light **lightList, int lightCount)
{
	Geometry::setLightList(lightList, lightCount);
}

// ----------------------------------------------------------------------

bool GL_wu2::setVertexList(const Vector *vertices, int vertexCount, const Vector *normal, int normalCount)
{
	return Geometry::setVertexList(vertices, vertexCount, normal, normalCount);
}

// ----------------------------------------------------------------------

void GL_wu2::setMaterial(const StaticMaterial *mat)
{
	if (mat == NULL)
	{
		material = NULL;
		return;
	}

	// see if we should ignore this polygon based on the material
	material = reinterpret_cast<const MaterialData_wu2 *>(mat->getData());
}

// ----------------------------------------------------------------------

#if 0
inline real triangleSize(const PolyVertex *v0, const PolyVertex *v1, const PolyVertex *v2)
{
	return CONST_REAL(0.5) * ((v2->x * v0->y) - (v2->y * v0->x) + (v0->x * v1->y) - (v0->y * v1->x) + (v1->x * v2->y) - (v1->y * v2->x));
}
#endif

// ----------------------------------------------------------------------

#if 0
void GL_wu2::polygon(const PolyVertex *polyVertex, int vertexCount, int flags)
{
	int                     i, j, k;
	raster::texture        *tmap;
	raster::poly           *p;
	float                   uv[3][2];
	const PolyVertex       *v0, *v1, *v2;
	bool                    affine, flat, transparency;
	float                   bestSize, newSize;
	raster::vert_3          tv[RASTER_VERTEX_CNT];
	raster::vert_3s         v[3];
	const raster::vert_3   *tvi[RASTER_VERTEX_CNT];

	
	DEBUG_FATAL(vertexCount < 3 || vertexCount > RASTER_VERTEX_CNT, ("GL_wu2::polygon %d vertices", vertexCount));

	if (!material || DFlag(DF_E_HIDE_POLYS))
		return;

	// handle wirefill specially
	if (DFlag(DF_E_WIRE_FILL))
	{
		raster::poly_outline p;
		raster::vert_2d      v[RASTER_VERTEX_CNT];

		// build all the raster 2d vertices, linking them counter-clockwise
		for (i = 0; i < vertexCount; ++i)
		{
			v[i].x    = REAL_TO_FLOAT(polyVertex[i].x - vx0);
			v[i].y    = REAL_TO_FLOAT(polyVertex[i].y - vy0);
			v[i].next = v+i-1;
		}
		v[0].next   = v+vertexCount-1;

		// set the material data
		p.thickness = 1;
		p.shade     = 1;
		p.surf      = wirefill->getSolidSurf();
		p.verts     = v;
		
		// draw the poly
		TRACE(("a"));
		raster::draw_poly(&p);
		TRACE(("b"));

		return;
	}

	// support for pete's cockpit gauge polygons
	if (flags & PF_BITMAP)
	{
		raster::poly_aft p;
		raster::vert_2d  rasterVertex[RASTER_VERTEX_CNT];

		for (i = 0; i < vertexCount; i++)
		{
			rasterVertex[i].x    = REAL_TO_FLOAT(polyVertex[i].x);
			rasterVertex[i].y    = REAL_TO_FLOAT(polyVertex[i].y);
			rasterVertex[i].next = rasterVertex+i-1;
		}
		rasterVertex[0].next   = rasterVertex+vertexCount-1;

		// set the polygon data
		p.tmap      = material->getTmap();

		p.verts     = rasterVertex;
		p.shade     = CONST_REAL(1.0);
		p.us[0]     = REAL_TO_FLOAT(polyVertex[0].u);
		p.us[1]     = REAL_TO_FLOAT(polyVertex[vertexCount-1].u);
		p.us[2]     = REAL_TO_FLOAT(polyVertex[vertexCount-2].u);
		p.vs[0]     = REAL_TO_FLOAT(polyVertex[0].v);
		p.vs[1]     = REAL_TO_FLOAT(polyVertex[vertexCount-1].v);
		p.vs[2]     = REAL_TO_FLOAT(polyVertex[vertexCount-2].v);

		// draw the poly
		raster::draw_poly(&p);

		return;
	}

	// get the three points that make the largest tri
	v0 = polyVertex+2;
	v1 = polyVertex+1;
	v2 = polyVertex+0;
	if (vertexCount > 3)
	{
		// look for the largest triangle to get the best mapping
		bestSize = -1;
		for (i =   0; i < vertexCount-2; ++i)
			for (j = i+1; j < vertexCount-1; ++j)
				for (k = j+1; k < vertexCount-0; ++k)
				{
					newSize = triangleSize(polyVertex+i, polyVertex+j, polyVertex+k);
					if (newSize > bestSize)
					{
						bestSize = newSize;
						v0 = polyVertex+k;
						v1 = polyVertex+j;
						v2 = polyVertex+i;
					}
				}
	}
	else
		bestSize = triangleSize(polyVertex+0, polyVertex+1, polyVertex+2);

	// setup the mapping vertices				
	v[0].c[0]  =  v0->x - vx0;
	v[0].c[1]  =  v0->y - vy0;
	v[0].z_inv =  v0->ooz;
	v[0].shade =  v0->b;

	v[1].c[0]  =  v1->x - vx0;
	v[1].c[1]  =  v1->y - vy0;
	v[1].z_inv =  v1->ooz;
	v[1].shade =  v1->b;

	v[2].c[0]  =  v2->x - vx0;
	v[2].c[1]  =  v2->y - vy0;
	v[2].z_inv =  v2->ooz;
	v[2].shade =  v2->b;

	// handle fully-lit screen polys appropriately
	if (flags & PF_FULLY_LIT)
	{
		v[0].shade = CONST_REAL(1);
		v[1].shade = CONST_REAL(1);
		v[2].shade = CONST_REAL(1);
	}	

	// debug flag to force flat shading
	flat = (flags & PF_FLAT_SHADE) != 0;
	if (DFlag(DF_E_FLAT_SHADE))
		flat = true;

	// determine if the polygon is flat enough, or small enough, to affine
	affine = (flags & PF_AFFINE) != 0;
	if (bestSize < CONST_REAL(300))
		affine = true;
	if (DFlag(DF_E_AFFINE))
		affine = true;
	if (DFlag(DF_E_PERSPECTIVE))
		affine = false;

	// get the material information
	tmap = material->getTmap();
	transparency = (tmap && tmap->has_transparency) ? true : false;

#if 0
	// -qq- work around bugs in the rasterizer
	transparency = false;
#endif

	// setup the uv coordinates for the mapping vertices
	if (tmap)
	{
		uv[0][0]   =  v0->u;
		uv[0][1]   =  v0->v;
		uv[1][0]   =  v1->u;
		uv[1][1]   =  v1->v;
		uv[2][0]   =  v2->u;
		uv[2][1]   =  v2->v;
	}

	// setup the vertex array for transparency primitives
	if (transparency)
		for (i = 0, j = vertexCount-1; i < vertexCount; ++i, --j)
		{
			tvi[i] = tv+i;
			tv[i].c[0]  =  polyVertex[j].x - vx0;
			tv[i].c[1]  =  polyVertex[j].y - vy0;
		}

	p = NULL;

	if (!tmap)
	{
		if (flat)
		{
			// solid flat
			p = submit_afs(v[0], v[1], v[2], polyVertex[0].b, material->getSolidSurf());
		}
		else
		{
			// solid gouraud
			p = submit_als(v[0], v[1], v[2], material->getSolidSurf());
		}
	}
	else
	{	
		if (affine)
		{
			++Camera::renderStats->polygonsAffine;

			if (flat)
			{
				if (transparency)
				{
					// affine flat transparent
					submit_aft_tp(v[0], v[1], v[2], polyVertex[0].b, tmap, uv[0], uv[1], uv[2], tvi, vertexCount);
				}
				else
				{
					// affine flat opaque
					p = submit_aft(v[0], v[1], v[2], polyVertex[0].b, tmap, uv[0], uv[1], uv[2]);
				}
			}
			else
			{
				if (transparency)
				{
					// affine gouraud transparent
					submit_alt_tp(v[0], v[1], v[2], tmap, uv[0], uv[1], uv[2], tvi, vertexCount);
				}
				else
				{
					// affine gouraud opaque
					p = submit_alt(v[0], v[1], v[2], tmap, uv[0], uv[1], uv[2]);
				}
			}
		}
		else
		{
			++Camera::renderStats->polygonsPerspective16;

			if (flat)
			{
				if (transparency)
				{
					// perspective flat transparent
					submit_pft_tp(v[0], v[1], v[2], polyVertex[0].b, tmap, uv[0], uv[1], uv[2], tvi, vertexCount);
				}
				else
				{
					// perspective flat opaque
					p = submit_pft(v[0], v[1], v[2], polyVertex[0].b, tmap, uv[0], uv[1], uv[2]);
				}
			}
			else
			{
				if (transparency)
				{
					// perspective gouraud transparent
					submit_plt_tp(v[0], v[1], v[2], tmap, uv[0], uv[1], uv[2], tvi, vertexCount);
				}
				else
				{
					// perspective gouraud flat
					p = submit_plt(v[0], v[1], v[2], tmap, uv[0], uv[1], uv[2]);
				}
			}
		}
	}

	// if we added a polygon, submit all its edges
	if (p)
	{
		for (i = vertexCount-1; i > 0; i--)
			submit_edge(polyVertex[i].x-vx0,polyVertex[i].y-vy0,polyVertex[i-1].x-vx0,polyVertex[i-1].y-vy0,p,NULL);
		submit_edge(polyVertex[0].x-vx0,polyVertex[0].y-vy0,polyVertex[vertexCount-1].x-vx0,polyVertex[vertexCount-1].y-vy0,p,NULL);
	}

	// in the case where we are not in a 3d scene, pop these polys into the frame buffer immediately
	if (!in3dRender)
	{
		sweep_and_draw();
		draw_transparent_polys();
	}
}

#endif

// ----------------------------------------------------------------------

void GL_wu2::drawPointList(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 1, ("Gl_wu2::drawPointListU %d vertices", numberOfVertices));
	Geometry::drawPointList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawLineList(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 2 || (numberOfVertices & 1), ("Gl_wu2::drawLineListU %d vertices", numberOfVertices));
	Geometry::drawLineList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawLineStrip(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 2, ("Gl_wu2::drawLineStripU %d vertices", numberOfVertices));
	Geometry::drawLineStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawTriangleList(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3 || (numberOfVertices % 3 != 0), ("Gl_wu2::drawTriangleListU %d vertices", numberOfVertices));
	Geometry::drawTriangleList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawTriangleStrip(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3, ("Gl_wu2::drawTriangleStripU %d vertices", numberOfVertices));
	Geometry::drawTriangleStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawTriangleFan(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3, ("Gl_wu2::drawTriangleFanU %d vertices", numberOfVertices));
	Geometry::drawTriangleFan(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawPointList(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 1, ("Gl_wu2::drawPointListT %d vertices", numberOfVertices));
	submitPointList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawLineList(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 2 || (numberOfVertices & 1), ("Gl_wu2::drawLineListT %d vertices", numberOfVertices));
	submitLineList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawLineStrip(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 2, ("Gl_wu2::drawLineStripT %d vertices", numberOfVertices));
	submitLineStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawTriangleList(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 3 || (numberOfVertices % 3 != 0), ("GL_wu2::drawTriangleListT %d vertices", numberOfVertices));
	submitTriangleList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawTriangleStrip(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 3, ("GL_wu2::drawTriangleStripT %d vertices", numberOfVertices));
	submitTriangleStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::drawTriangleFan(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 3, ("GL_wu2::drawTriangleFanT %d vertices", numberOfVertices));
	submitTriangleFan(vertexList, numberOfVertices);
}


// ----------------------------------------------------------------------

void GL_wu2::passPointList(const TVertex *vertexList, int numberOfVertices)
{
	UNREF(vertexList);
	UNREF(numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::passLineList(const TVertex *vertexList, int numberOfVertices)
{
	UNREF(vertexList);
	UNREF(numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::passLineStrip(const TVertex *vertexList, int numberOfVertices)
{
	UNREF(vertexList);
	UNREF(numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::passTriangleList(const TVertex *vertexList, int numberOfVertices)
{
	raster::vert_3s              v[3];
	float                        uv[3][2];
	raster::solid_surface_desc  *solid;
	raster::texture             *tmap;
	raster::poly                *p;

	tmap = material->getTexture();
	solid = material->getSolidSurfaceDesc();

	for ( ; numberOfVertices; numberOfVertices -= 3, vertexList += 3)
	{
		// setup the vertices
		v[0].c[0]  =  vertexList[0].x - vx0;
		v[0].c[1]  =  vertexList[0].y - vy0;
		v[0].z_inv =  vertexList[0].ooz;
		v[0].shade =  vertexList[0].b;

		v[1].c[0]  =  vertexList[1].x - vx0;
		v[1].c[1]  =  vertexList[1].y - vy0;
		v[1].z_inv =  vertexList[1].ooz;
		v[1].shade =  vertexList[1].b;

		v[2].c[0]  =  vertexList[2].x - vx0;
		v[2].c[1]  =  vertexList[2].y - vy0;
		v[2].z_inv =  vertexList[2].ooz;
		v[2].shade =  vertexList[2].b;

		// setup the texture mapping coordinates
		uv[0][0]   =  vertexList[0].u;
		uv[0][1]   =  vertexList[0].v;
		uv[1][0]   =  vertexList[1].u;
		uv[1][1]   =  vertexList[1].v;
		uv[2][0]   =  vertexList[2].u;
		uv[2][1]   =  vertexList[2].v;

		if (tmap)
			p = submit_plt(v[0], v[2], v[1], tmap, uv[0], uv[2], uv[1]);
		else
			if (solid)
				p = submit_pls(v[0], v[2], v[1], solid);
			else
				p = NULL;

		if (p)
		{
			submit_edge(v[0].c[0], v[0].c[1], v[2].c[0], v[2].c[1], p, NULL);
			submit_edge(v[1].c[0], v[1].c[1], v[0].c[0], v[0].c[1], p, NULL);
			submit_edge(v[2].c[0], v[2].c[1], v[1].c[0], v[1].c[1], p, NULL);
		}
	}

	// in the case where we are not in a 3d scene, pop these polys into the frame buffer immediately
	if (!in3dRender)
	{
		sweep_and_draw();
		draw_transparent_polys();
	}
}

// ----------------------------------------------------------------------

void GL_wu2::passTriangleStrip(const TVertex *vertexList, int numberOfVertices)
{
	UNREF(vertexList);
	UNREF(numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_wu2::passTriangleFan(const TVertex *vertexList, int numberOfVertices)
{
	typedef float uvt[2];

	raster::vert_3s              v[3];
	raster::vert_3s             *v0, *v1, *v2;
	float                        uv[3][2];
	uvt                         *uv0, *uv1, *uv2;
	raster::solid_surface_desc  *solid;
	raster::texture             *tmap;
	raster::poly                *p;

	tmap = material->getTexture();
	solid = material->getSolidSurfaceDesc();

	// setup vertex 0
	v0 = v+0;
	uv0 = uv+0;
	v[0].c[0]  =  vertexList[0].x - vx0;
	v[0].c[1]  =  vertexList[0].y - vy0;
	v[0].z_inv =  vertexList[0].ooz;
	v[0].shade =  vertexList[0].b;
	uv[0][0]   =  vertexList[0].u;
	uv[0][1]   =  vertexList[0].v;

	// setup vertex 1
	v1 = v+1;
	uv1 = uv+1;
	v[1].c[0]  =  vertexList[1].x - vx0;
	v[1].c[1]  =  vertexList[1].y - vy0;
	v[1].z_inv =  vertexList[1].ooz;
	v[1].shade =  vertexList[1].b;
	uv[1][0]   =  vertexList[1].u;
	uv[1][1]   =  vertexList[1].v;

	// setup for walking along the fan
	v2  = v+2;
	uv2 = uv+2;
	
	for (numberOfVertices -= 2, vertexList += 2; numberOfVertices; --numberOfVertices, ++vertexList)
	{
		v2->c[0]  = vertexList->x - vx0;
		v2->c[1]  = vertexList->y - vy0;
		v2->z_inv = vertexList->ooz;
		v2->shade = vertexList->b;

		// setup the texture mapping coordinates
		(*uv2)[0] = vertexList->u;
		(*uv2)[1] = vertexList->v;

		if (tmap)
			p = submit_plt(*v0, *v2, *v1, tmap, *uv0, *uv2, *uv1);
		else
			if (solid)
				p = submit_pls(*v0, *v2, *v1, solid);
			else
				p = NULL;

		if (p)
		{
			DPRINT(("*"));
			submit_edge(v1->c[0], v1->c[1], v0->c[0], v0->c[1], p, NULL);
			submit_edge(v2->c[0], v2->c[1], v1->c[0], v1->c[1], p, NULL);
			submit_edge(v0->c[0], v0->c[1], v2->c[0], v2->c[1], p, NULL);
		}

		// advance through the fan
		swap(v1, v2);
		swap(uv1, uv2);
	}

	// in the case where we are not in a 3d scene, pop these polys into the frame buffer immediately
	if (!in3dRender)
	{
		sweep_and_draw();
		draw_transparent_polys();
	}
}

// ======================================================================
